/*
Copyright 2021.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package controllers

import (
	"context"
	"fmt"
	"time"

	"gopkg.in/yaml.v3"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime"
	corev1alpha1 "libre.sh/controller/api/v1alpha1"
	ctrl "sigs.k8s.io/controller-runtime"
	"sigs.k8s.io/controller-runtime/pkg/client"
	"sigs.k8s.io/controller-runtime/pkg/log"
)

// LibreStackReconciler reconciles a LibreStack object
type LibreStackReconciler struct {
	client.Client
	Scheme *runtime.Scheme
	Config corev1alpha1.LibreControllerConfig
}

//+kubebuilder:rbac:groups=core.libre.sh,resources=librestacks,verbs=get;list;watch;create;update;patch;delete
//+kubebuilder:rbac:groups=core.libre.sh,resources=librestacks/status,verbs=get;update;patch
//+kubebuilder:rbac:groups=core.libre.sh,resources=librestacks/finalizers,verbs=update

// Reconcile is part of the main kubernetes reconciliation loop which aims to
// move the current state of the cluster closer to the desired state.
// TODO(user): Modify the Reconcile function to compare the state specified by
// the LibreStack object against the actual cluster state, and then
// perform operations to make the cluster state reflect the state specified by
// the user.
//
// For more details, check Reconcile and its Result here:
// - https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.8.3/pkg/reconcile
func (r *LibreStackReconciler) Reconcile(ctx context.Context, req ctrl.Request) (ctrl.Result, error) {
	log := log.FromContext(ctx)

	wrapper := StackWrapper{
		LibreStackReconciler: r,
		Applications:         make(map[string]corev1alpha1.LibreApplication),
	}
	if err := r.Get(ctx, req.NamespacedName, &wrapper.Stack); err != nil {
		log.Error(err, "unable to fetch LibreStack")
		return ctrl.Result{}, client.IgnoreNotFound(err)
	}

	if err := wrapper.CompileTemplate(ctx); err != nil {
		log.Error(err, "unable to compile LibreStack template")
		return ctrl.Result{}, client.IgnoreNotFound(err)
	}

	if err := wrapper.HydrateTemplate(ctx); err != nil {
		log.Error(err, "unable to hydrate LibreStack template")
		return ctrl.Result{}, err
	}

	if err := wrapper.LoadApplications(ctx); err != nil {
		return ctrl.Result{}, err
	}

	// Reconcil postgres
	if wrapper.IsPostgres() {
		postgres := unstructured.Unstructured{}
		wrapper.SetPostgresMeta(&postgres)
		op, err := ctrl.CreateOrUpdate(ctx, r.Client, &postgres, func() error {
			wrapper.SetPostgresSpec(&postgres)
			return nil
		})
		if err != nil {
			return ctrl.Result{}, err
		} else {
			log.Info("Posgres successfully reconciled", "operation", op)
		}
	}

	compiledTemplate := corev1.ConfigMap{
		ObjectMeta: metav1.ObjectMeta{
			Name:      fmt.Sprintf("%s-compiled-template", req.Name),
			Namespace: req.Namespace,
		},
		Data: make(map[string]string),
	}
	tplBytes, err := yaml.Marshal(wrapper.Template)
	if err != nil {
		return ctrl.Result{}, err
	}
	_, err = ctrl.CreateOrUpdate(ctx, r.Client, &compiledTemplate, func() error {
		compiledTemplate.Data["template"] = string(tplBytes)
		return nil
	})
	if err != nil {
		return ctrl.Result{}, err
	}

	return ctrl.Result{RequeueAfter: 10 * time.Second}, nil
}

// func (r *LibreStackReconciler) Config(ctx context.Context) corev1alpha1.LibreControllerConfig {
// 	var config corev1alpha1.LibreControllerConfig
// 	var cm corev1.ConfigMap
// 	if err := r.Get(ctx, req.NamespacedName, &cm); err != nil {
// 		log.Error(err, "unable to fetch LibreStack")
// 		return ctrl.Result{}, client.IgnoreNotFound(err)
// 	}
// 	return config
// }

// SetupWithManager sets up the controller with the Manager.
func (r *LibreStackReconciler) SetupWithManager(mgr ctrl.Manager) error {
	return ctrl.NewControllerManagedBy(mgr).
		For(&corev1alpha1.LibreStack{}).
		Complete(r)
}
