/*
Copyright 2021.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package controllers

import (
	"context"
	"encoding/json"
	"errors"

	jsonpatch "github.com/evanphx/json-patch"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/apimachinery/pkg/util/yaml"
	corev1alpha1 "libre.sh/controller/api/v1alpha1"
)

func (r *LibreStackReconciler) CompileStackTemplate(ctx context.Context, stack *corev1alpha1.LibreStack) (corev1alpha1.StackTemplate, error) {
	var tpl []byte

	refs := append(r.Config.BaseTemplatesFrom, stack.Spec.TemplatesFrom...)
	for _, v := range refs {
		t, err := r.GetTemplateFrom(ctx, stack.Namespace, v)
		if err != nil {
			return corev1alpha1.StackTemplate{}, err
		}
		if tpl == nil {
			tpl = t
			continue
		}
		tpl, err = jsonpatch.MergePatch(tpl, t)
		if err != nil {
			return corev1alpha1.StackTemplate{}, err
		}
	}

	if stack.Spec.Template != nil {
		stackTpl, err := json.Marshal(stack.Spec.Template)
		if err != nil {
			return corev1alpha1.StackTemplate{}, err
		}

		tpl, err = jsonpatch.MergePatch(tpl, stackTpl)
		if err != nil {
			return corev1alpha1.StackTemplate{}, err
		}
	}

	var output corev1alpha1.StackTemplate
	if err := json.Unmarshal(tpl, &output); err != nil {
		return corev1alpha1.StackTemplate{}, err
	}

	return output, nil
}

func (r *LibreStackReconciler) GetTemplateFrom(ctx context.Context, defaultNamespace string, ref corev1alpha1.StackTemplateReference) ([]byte, error) {
	var tpl []byte

	namespaceName := types.NamespacedName{
		Name:      ref.Name,
		Namespace: ref.Namespace,
	}
	if namespaceName.Namespace == "" {
		namespaceName.Namespace = defaultNamespace
	}

	switch ref.Kind {
	case "ConfigMap":
		var cm corev1.ConfigMap
		if err := r.Get(ctx, namespaceName, &cm); err != nil {
			return tpl, err
		}
		if val, ok := cm.Data["template"]; ok {
			return yaml.ToJSON([]byte(val))
		}
	case "Secret":
		var secret corev1.Secret
		if err := r.Get(ctx, namespaceName, &secret); err != nil {
			return tpl, err
		}
		if val, ok := secret.StringData["template"]; ok {
			return yaml.ToJSON([]byte(val))
		}
	}

	return nil, errors.New("template not found")
}

// func (r *LibreStackReconciler) HydrateStackTemplate(ctx context.Context, stack *corev1alpha1.LibreStack, tpl *corev1alpha1.StackTemplate) error {
// 	wrapper := StackWrapper{
// 		Stack:        stack,
// 		Template:     tpl,
// 		Applications: make(map[string]corev1alpha1.LibreApplication),
// 	}
// 	for appName, appTpl := range tpl.Applications {
// 		if appTpl.Parameters.Domain == "" {
// 			appTpl.Parameters.Domain = fmt.Sprintf("%s.%s", appName, tpl.Domain)
// 		}

// 		if appTpl.Application.Namespace == "" {
// 			appTpl.Application.Namespace = stack.Namespace
// 		}

// 		for aN, a := range tpl.Applications {
// 			var application corev1alpha1.LibreApplication
// 			if err := r.Client.Get(ctx, types.NamespacedName(a.Application), &application); err != nil {
// 				return err
// 			}

// 			wrapper.Applications[aN] = application
// 		}

// 		tpl.Applications[appName] = appTpl
// 	}
// 	return nil
// }
