/*
Copyright 2021.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package controllers

import (
	"context"
	"encoding/json"
	"errors"
	"fmt"

	jsonpatch "github.com/evanphx/json-patch"
	corev1 "k8s.io/api/core/v1"
	"k8s.io/apimachinery/pkg/types"
	"k8s.io/apimachinery/pkg/util/yaml"
	corev1alpha1 "libre.sh/controller/api/v1alpha1"
)

type StackWrapper struct {
	*LibreStackReconciler
	Stack        corev1alpha1.LibreStack
	Template     corev1alpha1.StackTemplate
	Applications map[string]corev1alpha1.LibreApplication
}

func (w *StackWrapper) ApplicationTemplate(name string) corev1alpha1.ApplicationTemplate {
	return w.Template.Applications[name]
}

func (w *StackWrapper) Application(name string) corev1alpha1.LibreApplication {
	return w.Applications[name]
}

func (w *StackWrapper) CompileTemplate(ctx context.Context) error {
	var tpl []byte

	refs := append(w.Config.BaseTemplatesFrom, w.Stack.Spec.TemplatesFrom...)
	for _, v := range refs {
		t, err := w.GetTemplateFrom(ctx, w.Stack.Namespace, v)
		if err != nil {
			return err
		}
		if tpl == nil {
			tpl = t
			continue
		}
		tpl, err = jsonpatch.MergePatch(tpl, t)
		if err != nil {
			return err
		}
	}

	if w.Stack.Spec.Template != nil {
		stackTpl, err := json.Marshal(w.Stack.Spec.Template)
		if err != nil {
			return err
		}

		tpl, err = jsonpatch.MergePatch(tpl, stackTpl)
		if err != nil {
			return err
		}
	}

	if err := json.Unmarshal(tpl, &w.Template); err != nil {
		return err
	}

	return nil
}

func (w *StackWrapper) GetTemplateFrom(ctx context.Context, defaultNamespace string, ref corev1alpha1.StackTemplateReference) ([]byte, error) {
	var tpl []byte

	namespaceName := types.NamespacedName{
		Name:      ref.Name,
		Namespace: ref.Namespace,
	}
	if namespaceName.Namespace == "" {
		namespaceName.Namespace = defaultNamespace
	}

	switch ref.Kind {
	case "ConfigMap":
		var cm corev1.ConfigMap
		if err := w.Get(ctx, namespaceName, &cm); err != nil {
			return tpl, err
		}
		if val, ok := cm.Data["template"]; ok {
			return yaml.ToJSON([]byte(val))
		}
	case "Secret":
		var secret corev1.Secret
		if err := w.Get(ctx, namespaceName, &secret); err != nil {
			return tpl, err
		}
		if val, ok := secret.StringData["template"]; ok {
			return yaml.ToJSON([]byte(val))
		}
	}

	return nil, errors.New("template not found")
}

func (w *StackWrapper) HydrateTemplate(ctx context.Context) error {
	if w.Template.Postgres.CPU == "" {
		w.Template.Postgres.CPU = "1"
	}

	if w.Template.Postgres.Memory == "" {
		w.Template.Postgres.Memory = "500Mi"
	}

	if w.Template.Postgres.Size == "" {
		w.Template.Postgres.Size = "5G"
	}

	for appName, appTpl := range w.Template.Applications {
		if appTpl.Parameters.Domain == "" {
			appTpl.Parameters.Domain = fmt.Sprintf("%s.%s", appName, w.Template.Domain)
		}

		if appTpl.Application.Namespace == "" {
			appTpl.Application.Namespace = w.Stack.Namespace
		}

		w.Template.Applications[appName] = appTpl
	}
	return nil
}

func (w *StackWrapper) LoadApplications(ctx context.Context) error {
	for aN, a := range w.Template.Applications {
		var application corev1alpha1.LibreApplication
		if err := w.Client.Get(ctx, types.NamespacedName(a.Application), &application); err != nil {
			return err
		}
		w.Applications[aN] = application
	}
	return nil
}
